#include <assert.h>
#include <malloc.h>
#include <stdio.h>

#include "libpdf/pdfbool.h"
#include "libpdf/pdfdictionary.h"
#include "libpdf/pdfint.h"
#include "libpdf/pdffloat.h"
#include "libpdf/pdfliteral.h"
#include "libpdf/pdfobject.h"
/*#include "libpdf/pdfname.h"*/
#include "libpdf/pdfqueue.h"
#include "libpdf/pdfstream.h"
#include "libpdf/pdfstring.h"

/* note that there is no generic object allocator.  This function is
   called to allocate and set up the PDF header for an object, by that
   object's custom _new routine.  */

PDF_Object PDF_Object_create(int size, PDF_ObjType type, PDF_CBool indirect)
{
	PDF_Object retval;

	/* allocate space and set up header */
	retval = calloc(size, 1);
	
	if (retval == NULL)
		return NULL;

	PDF_Header_init((PDF_Header) retval, type, indirect);

	return retval;
}

/* delete a PDF object.  Call the customized destructor to get rid of
   the custom fields for each object; then get rid of the header, and
   finally get rid of the object itself */
void PDF_Object_delete(PDF_Object obj)
{
	switch (obj->header.objtype) {
		/* objects lacking internal structure don't need custom
		   destructors */
	  case PDF_INT:
		break;

	  case PDF_LITERAL:
		PDF_Literal_delete((PDF_Literal) obj);
		break;

	  case PDF_DICT_ENTRY:
		PDF_Dict_Entry_delete((PDF_Dict_Entry) obj);
		break;
		
	  case PDF_STREAM:
		PDF_Stream_delete((PDF_Stream) obj);
		break;
	}
	
	free(obj);
}

void PDF_Object_set_indirect(PDF_Object obj, PDF_CBool flag)
{
	obj->header.indirect = flag;
}

PDF_CBool PDF_Object_get_indirect(PDF_Object obj)
{
	return obj->header.indirect;
}

int PDF_Object_write(PDF_Object obj, FILE *ostream, int index,
					 PDF_CBool write_indirect, PDF_Xref_Table xref,
					 PDF_Queue indqueue)
{
	int numwritten;
	int pos;

	/* If we've got an indirect object, and it doesn't have an object
	   number, we need to assign it one.  If this isn't the top level
	   queue processing, we also need to put it in the output queue so
	   it'll be written later. */
	if (obj->header.indirect && (obj->header.objnum == -1)) {
		obj->header.objnum = PDF_Xref_Table_new_objnum(xref);
		if (!write_indirect)
			PDF_Queue_append(indqueue, obj);
	}
		
	/* it turns out that the write_indirect flag is only turned on
	   when this is called from the top level (when the top-level
	   document is being processed), and that the only objects at that
	   level are indirect.  So if the write_indirect flag is on, we
	   need to generate a cross-reference and print the object header */
	if (write_indirect) {
		assert(obj->header.indirect);
		
		PDF_Xref_Table_append(xref, obj->header.objnum, index,
							  obj->header.version);

		numwritten = fprintf(ostream, "%d %d obj\n",
							 obj->header.objnum, obj->header.version);
		
		if (numwritten < 0)
			return -1;
		index += numwritten;
	}

	/* if either write_indirect is on, or the object isn't indirect,
	   we need to output the object.  In either case, if this is a
	   compound object, we call recursively but don't turn on
	   write_indirect for recursive calls.

	   If write_indirect is off, and this is an indirect object, print
	   a reference to it.
	*/
	if (write_indirect || !obj->header.indirect) {
		switch (obj->header.objtype) {
		  case PDF_INT:
			index = PDF_Int_write((PDF_Int) obj, ostream, index,
								  PDF_CBOOL_FALSE, xref, indqueue);
			break;
			
		  case PDF_DICT_ENTRY:
			index = PDF_Dict_Entry_write((PDF_Dict_Entry) obj, ostream, index,
										 PDF_CBOOL_FALSE, xref, indqueue);
			break;

		  case PDF_LITERAL:
			index = PDF_Literal_write((PDF_Literal) obj, ostream, index,
									  PDF_CBOOL_FALSE, xref, indqueue);
			break;

		  case PDF_QUEUE:
			index = PDF_Queue_write((PDF_Queue) obj, ostream, index,
									PDF_CBOOL_FALSE, xref, indqueue);
			break;
			
		  case PDF_STREAM:
			index = PDF_Stream_write((PDF_Stream) obj, ostream, index,
									 PDF_CBOOL_FALSE, xref, indqueue);
			break;
		}
		if (index < 0)
			return -1;
	} else {
			numwritten = fprintf(ostream, "%d %d R",
								 obj->header.objnum, obj->header.version);
			if (numwritten < 0)
				return -1;
			index += numwritten;
	}

	/* If this is an indirect object we've written, we need to write the
	   trailing endobj */
	if (obj->header.indirect && write_indirect) {
		numwritten = fprintf(ostream, "\nendobj\n");
		if (numwritten < 0)
			return -1;
		index += numwritten;
	}
	
	return index;
}
